/*
 * Copyright (c) 2001-2006, John Mettraux, OpenWFE.org
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * . Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.  
 * 
 * . Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * 
 * . Neither the name of the "OpenWFE" nor the names of its contributors may be
 *   used to endorse or promote products derived from this software without
 *   specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: XlobSession.java 3149 2006-08-31 04:48:48Z jmettraux $
 */

//
// XlobSession.java
//
// john.mettraux@openwfe.org
//
// generated with 
// jtmpl 1.1.01 2004/05/19 (john.mettraux@openwfe.org)
//

package openwfe.org.xlob;

import openwfe.org.sql.ds.OwfeDataSource;
import openwfe.org.engine.expressions.FlowExpressionId;


/**
 * Xlob is an OpenWFE library for storing XML versions of expressions
 * and workitems in a database CLOB.
 *
 * <p><font size=2>CVS Info :
 * <br>$Author: jmettraux $
 * <br>$Id: XlobSession.java 3149 2006-08-31 04:48:48Z jmettraux $ </font>
 *
 * @author john.mettraux@openwfe.org
 */
public interface XlobSession
{

    /**
     * Inits the session (once it got instantiated).
     *
     * @param prefix each OpenWFE using an xlob session must explicitely
     * specify a prefix like 'worklist' or 'engine-alpha'.
     */
    public void init (java.sql.Connection con, String prefix);

    /**
     * Inits the session (once it got instantiated).
     *
     * @param prefix each OpenWFE using an xlob session must explicitely
     * specify a prefix like 'worklist' or 'engine-alpha'.
     */
    public void init (OwfeDataSource ds, String prefix);

    /**
     * Closes this session.
     */
    public void close ();

    /**
     * Stores an XML document for a given fei.
     */
    public void store (FlowExpressionId fei, org.jdom.Document doc)
        throws XlobException;

    /**
     * Stores an XML document for a given fei.
     */
    public void store (FlowExpressionId fei, org.jdom.Element elt)
        throws XlobException;

    /**
     * Stores an XML document for a given fei.
     */
    public void store (FlowExpressionId fei, String sXml)
        throws XlobException;

    /**
     * Turns a bean into an XML document (openwfe.org.xml.XmlCoder) and
     * stores it.
     */
    public void storeBean (FlowExpressionId fei, Object bean)
        throws XlobException;

    /**
     * Updates or inserts an XML document for a given fei.
     */
    public void remove (FlowExpressionId fei)
        throws XlobException;

    /**
     * Finds the document matching precisely the given fei.
     */
    public org.jdom.Document find (FlowExpressionId fei)
        throws XlobException;

    /**
     * Retrieves the corresponding XML doc and turns it into its original
     * bean (openwfe.org.xml.XmlCoder).
     */
    public Object findBean (FlowExpressionId fei)
        throws XlobException;

    /**
     * Will return true if the given fei points to a stored document.
     */
    public boolean isStored (FlowExpressionId fei)
        throws XlobException;

    /**
     * Given a partial fei (not all fields set), returns a set of
     * matching xml  documents.
     * The fei fields may contain SQL wildcards...
     */
    public java.util.Set query (FlowExpressionId fei)
        throws XlobException;

    /**
     * Given a partial fei (not all fields set), returns a set of
     * matching xml  documents, but no more than a limit count.
     * The fei fields may contain SQL wildcards...
     */
    public java.util.Set query (FlowExpressionId fei, int limit)
        throws XlobException;

    /**
     * Given a partial fei (not all fields set), returns a set of
     * matching workitems.
     * The fei fields may contain SQL wildcards...
     */
    public java.util.Set queryBeans (FlowExpressionId fei)
        throws XlobException;

    /**
     * Given a partial fei (not all fields set), returns a set of
     * matching workitems, but no more than 'limit' count.
     * The fei fields may contain SQL wildcards...
     */
    public java.util.Set queryBeans (FlowExpressionId fei, int limit)
        throws XlobException;

    /**
     * Returns a list of all the flow expression ids.
     */
    public java.util.List listExpressionIds ()
        throws XlobException;

    /**
     * Given a partial fei returns the count of matching docs.
     * The fei fields may contain SQL wildcards...
     */
    public int count (FlowExpressionId fei)
        throws XlobException;

    /**
     * Returns the prefix used by this session.
     */
    public String getPrefix ();

    /**
     * Lists all prefixes (including the one for this 
     * session) used in the current db. 
     */
    public java.util.List listPrefixes ()
        throws XlobException;

}
