/*
 * Copyright (c) 2001-2006, John Mettraux, OpenWFE.org
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * . Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.  
 * 
 * . Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * 
 * . Neither the name of the "OpenWFE" nor the names of its contributors may be
 *   used to endorse or promote products derived from this software without
 *   specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: XlobSessionImpl.java 3310 2006-09-14 14:21:14Z jmettraux $
 */

//
// XlobSessionImpl.java
//
// john.mettraux@epfl.ch
//
// generated with 
// jtmpl 1.1.01 2004/05/19 (john.mettraux@openwfe.org)
//

package openwfe.org.xlob;

import openwfe.org.Utils;
import openwfe.org.MapUtils;
import openwfe.org.ServiceException;
import openwfe.org.ApplicationContext;
import openwfe.org.xml.XmlUtils;
import openwfe.org.sql.SqlUtils;
import openwfe.org.sql.ds.OwfeDataSource;
import openwfe.org.util.beancoder.XmlBeanCoder;
import openwfe.org.engine.expressions.FlowExpressionId;


/**
 * An implementation of XlobSession
 *
 * <p><font size=2>CVS Info :
 * <br>$Author: jmettraux $
 * <br>$Id: XlobSessionImpl.java 3310 2006-09-14 14:21:14Z jmettraux $ </font>
 *
 * @author john.mettraux@epfl.ch
 */
public class XlobSessionImpl

    implements XlobSession

{

    private final static org.apache.log4j.Logger log = org.apache.log4j.Logger
        .getLogger(XlobSessionImpl.class.getName());

    //
    // CONSTANTS & co

    private final static String FEI_FIELDS =
        "prefix, "+
        "engineId, "+
        "initialEngineId, "+
        "wfdUrl, "+
        "wfdName, "+
        "wfdRevision, "+
        "wfInstanceId, "+
        "expressionName, "+
        "expressionId";

    private final static String FIELDS =
        FEI_FIELDS + ", xldoc";

    /**
     * The parameter 'xlobPrefix'.
     */
    public final static String P_PREFIX
        = "xlobPrefix";

    //
    // FIELDS

    private java.sql.Connection connection = null;
    private OwfeDataSource dataSource = null;

    private String prefix = null;

    //
    // CONSTRUCTORS

    //
    // METHODS

    private java.sql.Connection getConnection ()
        throws java.sql.SQLException
    {
        if (this.connection != null) return this.connection;

        return this.dataSource.getConnection();
    }

    private void releaseConnection ()
    {
        if (this.dataSource != null) this.dataSource.releaseConnection();
    }

    private java.sql.PreparedStatement prepareStatement
        (final String header, FlowExpressionId fei)
    throws 
        java.sql.SQLException
    {
        log.debug("prepareStatement() for "+header+" / "+fei);

        if (fei == null) fei = new FlowExpressionId();

        final StringBuffer sb = new StringBuffer();

        sb
            .append(header)
            .append(" FROM xldocument")

            .append(" WHERE")
            .append(" prefix = ?");

        if (fei.getEngineId() != null)
            sb.append(" AND engineId = ?");

        if (fei.getInitialEngineId() != null)
            sb.append(" AND initialEngineId = ?");

        if (fei.getWorkflowDefinitionUrl() != null)
            sb.append(" AND wfdUrl = ?");

        if (fei.getWorkflowDefinitionName() != null)
            sb.append(" AND wfdName = ?");

        if (fei.getWorkflowDefinitionRevision() != null)
            sb.append(" AND wfdRevision = ?");

        if (fei.getWorkflowInstanceId() != null)
            sb.append(" AND wfInstanceId = ?");

        if (fei.getExpressionName() != null)
            sb.append(" AND expressionName = ?");

        if (fei.getExpressionId() != null)
            sb.append(" AND expressionId = ?");

        //
        // prepare statement

        final java.sql.PreparedStatement ps = 
            getConnection().prepareStatement(sb.toString());

        int i = 1;

        ps.setString(i++, this.prefix);

        if (fei.getEngineId() != null)
            ps.setString(i++, fei.getEngineId());

        if (fei.getInitialEngineId() != null)
            ps.setString(i++, fei.getInitialEngineId());

        if (fei.getWorkflowDefinitionUrl() != null)
            ps.setString(i++, fei.getWorkflowDefinitionUrl());

        if (fei.getWorkflowDefinitionName() != null)
            ps.setString(i++, fei.getWorkflowDefinitionName());

        if (fei.getWorkflowDefinitionRevision() != null)
            ps.setString(i++, fei.getWorkflowDefinitionRevision());

        if (fei.getWorkflowInstanceId() != null)
            ps.setString(i++, fei.getWorkflowInstanceId());

        if (fei.getExpressionName() != null)
            ps.setString(i++, fei.getExpressionName());

        if (fei.getExpressionId() != null)
            ps.setString(i++, fei.getExpressionId());

        return ps;
    }

    private QueryData doQuery 
        (String selectable, 
         final FlowExpressionId fei)
    throws 
        java.sql.SQLException, XlobException
    {
        if (selectable == null) selectable = "xldoc";

        log.debug("doQuery() for "+selectable+" / "+fei);

        final java.sql.PreparedStatement ps = 
            prepareStatement("SELECT "+selectable, fei);

        final java.sql.ResultSet rs = ps.executeQuery();

        return new QueryData(ps, rs);
    }

    private org.jdom.Document extract (final java.sql.ResultSet rs)
        throws XlobException
    {
        try
        {
            final org.jdom.input.SAXBuilder builder = 
                new org.jdom.input.SAXBuilder();

            if (getConnection().getClass().getName()
                    .startsWith("org.postgresql."))
            {
                final String sDoc = rs.getString(1);

                //log.debug("extract() doc is\n"+sDoc);

                return builder.build(new java.io.StringReader(sDoc));
            }

            final java.sql.Clob clob = rs.getClob(1);

            return builder.build(clob.getCharacterStream());
        }
        catch (final Throwable t)
        {
            throw new XlobException("extraction failure", t);
        }
    }

    private Object extractBean (final java.sql.ResultSet rs)
        throws XlobException
    {
        try
        {
            final org.jdom.Document d = extract(rs);

            return XmlBeanCoder.xmlDecode(d);
        }
        catch (final Throwable t)
        {
            throw new XlobException
                ("Failed to decode bean", t);
        }
    }

    private FlowExpressionId extractExpressionId (final java.sql.ResultSet rs)
        throws java.sql.SQLException
    {
        int i = 1;

        final FlowExpressionId fei = new FlowExpressionId();

        fei.setEngineId(rs.getString(i++));
        fei.setInitialEngineId(rs.getString(i++));
        fei.setWorkflowDefinitionUrl(rs.getString(i++));
        fei.setWorkflowDefinitionName(rs.getString(i++));
        fei.setWorkflowDefinitionRevision(rs.getString(i++));
        fei.setWorkflowInstanceId(rs.getString(i++));
        fei.setExpressionName(rs.getString(i++));
        fei.setExpressionId(rs.getString(i++));

        return fei;
    }

    //
    // METHODS from XlobSession

    /**
     * Inits the session (once it got instantiated).
     *
     * @param prefix each OpenWFE using an xlob session must explicitely
     * specify a prefix like 'worklist' or 'engine-alpha'.
     */
    public void init (final java.sql.Connection con, final String prefix)
    {
        this.connection = con;
        this.prefix = prefix;
    }

    /**
     * Inits the session (once it got instantiated).
     *
     * @param prefix each OpenWFE using an xlob session must explicitely
     * specify a prefix like 'worklist' or 'engine-alpha'.
     */
    public void init (OwfeDataSource ds, String prefix)
    {
        this.dataSource = ds;
        this.prefix = prefix;
    }

    /**
     * Closes this session.
     */
    public void close ()
    {
        try
        {
            if (this.connection != null) this.connection.close();
        }
        catch (final Throwable t)
        {
            log.info("close() problem", t);
        }
    }

    /*
     * closes the QueryData instance.
     */
    private void close (final QueryData qd)
    {
        if (qd == null) return;
        qd.close();
    }

    /**
     * Updates or inserts an XML document for a given fei.
     */
    public void store (final FlowExpressionId fei, final org.jdom.Document doc)
        throws XlobException
    {
        store(fei, XmlUtils.toString(doc));
    }

    /**
     * Updates or inserts an XML document for a given fei.
     */
    public void store (final FlowExpressionId fei, final org.jdom.Element elt)
        throws XlobException
    {
        store(fei, XmlUtils.toString(elt));
    }

    /**
     * Turns a bean into an XML document 
     * (openwfe.org.util.beancoder.XmlBeanCoder) and stores it.
     */
    public void storeBean (final FlowExpressionId fei, final Object bean)
        throws XlobException
    {
        try
        {
            store(fei, XmlBeanCoder.xmlEncode(bean));
        }
        catch (final Throwable t)
        {
            throw new XlobException("Failed to encode bean", t);
        }
    }

    /**
     * Updates or inserts an XML document for a given fei.
     */
    public void store (final FlowExpressionId fei, final String sXml)
        throws XlobException
    {
        java.sql.PreparedStatement ps = null;
        try
        {
            if (isStored(fei)) remove(fei);

            //
            // insert the doc

            final StringBuffer sb = new StringBuffer();

            sb
                .append("INSERT INTO xldocument ")
                .append("(")
                .append(FIELDS)
                .append(") VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");

            ps = getConnection().prepareStatement(sb.toString());

            int i = 1;

            ps.setString(i++, this.prefix);
            ps.setString(i++, fei.getEngineId());
            ps.setString(i++, fei.getInitialEngineId());
            ps.setString(i++, fei.getWorkflowDefinitionUrl());
            ps.setString(i++, fei.getWorkflowDefinitionName());
            ps.setString(i++, fei.getWorkflowDefinitionRevision());
            ps.setString(i++, fei.getWorkflowInstanceId());
            ps.setString(i++, fei.getExpressionName());
            ps.setString(i++, fei.getExpressionId());

            final java.io.StringReader reader = 
                new java.io.StringReader(sXml);

            ps.setCharacterStream(i++, reader, sXml.length());

            ps.executeUpdate();
        }
        catch (final Exception e)
        {
            throw new XlobException
                ("update failure for "+fei, e);
        }
        finally 
        {
            SqlUtils.closeStatement(ps);
            releaseConnection();
        }
    }

    /**
     * Updates or inserts an XML document for a given fei.
     */
    public void remove (final FlowExpressionId fei)
        throws XlobException
    {
        java.sql.PreparedStatement ps = null;

        try
        {
            ps = prepareStatement("DELETE ", fei);

            ps.executeUpdate();
        }
        catch (final Throwable t)
        {
            throw new XlobException("removal failure", t);
        }
        finally
        {
            SqlUtils.closeStatement(ps);
            releaseConnection();
        }
    }

    /**
     * Finds the document matching precisely the given fei.
     */
    public org.jdom.Document find (final FlowExpressionId fei)
        throws XlobException
    {
        QueryData qd = null;

        try
        {
            qd = doQuery(null, fei);

            qd.toNext();

            return extract(qd.getResultSet());
        }
        catch (final java.sql.SQLException e)
        {
            throw new XlobException
                ("Failed to find "+fei, e);
        }
        finally
        {
            close(qd);
            releaseConnection();
        }
    }

    /**
     * Retrieves the corresponding XML doc and turns it into its original
     * bean (openwfe.org.util.beancoder.XmlBeanCoder).
     */
    public Object findBean (final FlowExpressionId fei)
        throws XlobException
    {
        try
        {
            return XmlBeanCoder.xmlDecode(find(fei).getRootElement());
        }
        catch (final Throwable t)
        {
            throw new XlobException("decoding failure", t);
        }
    }

    /**
     * Will return true if the given fei points to a stored document.
     */
    public boolean isStored (final FlowExpressionId fei)
        throws XlobException
    {
        QueryData qd = null;

        try
        {
            qd = doQuery("count(*)", fei);

            if ( ! qd.toNext()) return false;

            return qd.getResultSet().getInt(1) > 0;
        }
        catch (final java.sql.SQLException se)
        {
            throw new XlobException("failure", se);
        }
        finally
        {
            close(qd);
            releaseConnection();
        }
    }

    /**
     * Given a partial fei (not all fields set), returns a set of
     * matching workitems.
     * The fei fields may contain SQL wildcards...
     */
    public java.util.Set query (final FlowExpressionId fei)
        throws XlobException
    {
        return query(fei, -1);
    }

    /**
     * Given a partial fei (not all fields set), returns a set of
     * matching xml  documents, but no more than a limit count.
     * The fei fields may contain SQL wildcards...
     */
    public java.util.Set query (FlowExpressionId fei, int limit)
        throws XlobException
    {
        QueryData qd = null;

        try
        {
            qd = doQuery(null, fei);

            final java.util.Set set = new java.util.HashSet();

            int count = 0;

            while (qd.toNext())
            {
                if (limit > 0 && count >= limit) break;

                set.add(extract(qd.getResultSet()));

                count++;
            }

            return set;
        }
        catch (final java.sql.SQLException e)
        {
            throw new XlobException 
                ("Failed to query for "+fei, e);
        }
        finally
        {
            close(qd);
            releaseConnection();
        }
    }

    /**
     * Given a partial fei (not all fields set), returns a set of
     * matching workitems.
     * The fei fields may contain SQL wildcards...
     */
    public java.util.Set queryBeans (final FlowExpressionId fei)
        throws XlobException
    {
        return queryBeans(fei, -1);
    }

    /**
     * Given a partial fei (not all fields set), returns a set of
     * matching workitems, but no more than 'limit' count.
     * The fei fields may contain SQL wildcards...
     */
    public java.util.Set queryBeans 
        (final FlowExpressionId fei, final int limit)
    throws 
        XlobException
    {
        QueryData qd = null;

        try
        {
            qd = doQuery(null, fei);

            final java.util.Set set = new java.util.HashSet();

            int count = 0;

            while (qd.toNext())
            {
                if (limit > 0 && count >= limit) break;

                set.add(extractBean(qd.getResultSet()));

                count++;
            }

            return set;
        }
        catch (final java.sql.SQLException e)
        {
            throw new XlobException 
                ("queryBeans() failure", e);
        }
        finally
        {
            close(qd);
            releaseConnection();
        }
    }

    /**
     * Given a partial fei returns the count of matching docs.
     * The fei fields may contain SQL wildcards... If fei is null,
     * will return the total count.
     */
    public int count (FlowExpressionId fei)
        throws XlobException
    {
        if (fei == null) fei = new FlowExpressionId();

        QueryData qd = null;

        try
        {
            qd = doQuery("count(*)", fei);

            qd.toNext();

            return qd.getResultSet().getInt(1);
        }
        catch (final java.sql.SQLException se)
        {
            throw new XlobException("count failure", se);
        }
        finally
        {
            close(qd);
            releaseConnection();
        }
    }

    /**
     * Returns a list of all the flow expression ids.
     */
    public java.util.List listExpressionIds ()
        throws XlobException
    {
        final StringBuffer sb = new StringBuffer();

        sb
            .append("SELECT ")
            .append(FEI_FIELDS)
            .append(" FROM xldocument")
            .append(" WHERE prefix = ?");

        java.sql.PreparedStatement st = null;
        java.sql.ResultSet rs = null;

        try
        {
            final java.util.List result = new java.util.ArrayList(50);

            st = getConnection().prepareStatement(sb.toString());

            st.setString(1, getPrefix());

            rs = st.executeQuery();

            while (rs.next())
                result.add(extractExpressionId(rs));

            return result;
        }
        catch (final java.sql.SQLException e)
        {
            throw new XlobException
                ("failed to list expression ids", e);
        }
        finally
        {
            SqlUtils.closeStatement(st, rs);
            releaseConnection();
        }
    }

    /**
     * Returns the prefix used by this session.
     */
    public String getPrefix ()
    {
        return this.prefix;
    }

    /**
     * Lists all prefixes (including the one for this 
     * session) used in the current db. 
     */
    public java.util.List listPrefixes ()
        throws XlobException
    {
        java.sql.Statement st = null;
        java.sql.ResultSet rs = null;

        try
        {
            st = getConnection().createStatement();

            rs = st.executeQuery("SELECT prefix FROM xldocument");

            final java.util.List l = new java.util.ArrayList(20);

            while (rs.next())
            {
                l.add(rs.getString(1));
            }

            return l;
        }
        catch (final java.sql.SQLException se)
        {
            throw new XlobException
                ("failed to list prefixes", se);
        }
        finally
        {
            SqlUtils.closeStatement(st, rs);
            releaseConnection();
        }
    }

    //
    // STATIC METHODS

    /**
     * A shortcut method for getting an XlobSession.
     */
    public static XlobSession getXlobSession 
        (final java.sql.Connection con, final String prefix)
    {
        final XlobSession s = new XlobSessionImpl();
        s.init(con, prefix);

        return s;
    }

    /**
     * A shortcut method for getting an XlobSession.
     */
    public static XlobSession getXlobSession 
        (final OwfeDataSource ds, final String prefix)
    {
        final XlobSession s = new XlobSessionImpl();
        s.init(ds, prefix);

        return s;
    }

    /**
     * Buils an XlobSession according to the instructions found
     * in the context and in the service params.
     */
    public static XlobSession createSession
        (final ApplicationContext context,
         final java.util.Map serviceParams)
    throws
        ServiceException
    {
        final OwfeDataSource ds = 
            SqlUtils.lookupDataSource(context, serviceParams);

        final String prefix = MapUtils
            .getAsString(serviceParams, P_PREFIX, context.getName());

        return getXlobSession(ds, prefix);
    }

    /**
     * Buils an XlobSession according to the instructions found
     * in the context and in the service params.
     */
    public static XlobSession createSession
        (final String prefix,
         final ApplicationContext context,
         final java.util.Map serviceParams)
    throws
        ServiceException
    {
        final OwfeDataSource ds = 
            SqlUtils.lookupDataSource(context, serviceParams);

        return getXlobSession(ds, prefix);
    }

    //
    // INNER CLASSES

    private static class QueryData
    {
        private java.sql.Statement st = null;
        private java.sql.ResultSet rs = null;

        public QueryData 
            (final java.sql.Statement st, final java.sql.ResultSet rs)
        {
            this.st = st;
            this.rs = rs;
        }

        public java.sql.ResultSet getResultSet ()
        {
            return this.rs;
        }

        public void close ()
        {
            SqlUtils.closeStatement(st, rs);
        }

        public boolean toNext ()
            throws XlobException
        {
            try
            {
                return this.rs.next();
            }
            catch (final Throwable t)
            {
                throw new XlobException("Failed to fetch next row", t);
            }
        }
    }

}
